import processing.core.*; 
import processing.xml.*; 

import controlP5.*; 
import processing.pdf.*; 
import java.awt.event.ComponentEvent; 
import java.awt.event.ComponentAdapter; 
import java.awt.Insets; 
import java.awt.Rectangle; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class DeadLeaves_Generator extends PApplet {








ControlP5 controlP5;
ControlWindow controlWindow;

boolean background = true;
boolean exponential = true;
boolean seamless = true;
boolean sorted = true;

boolean doSavePDF = false; // for PDF export

int nbObj = 512;  // initial number of objects
int type;         // to memorize the type of objects

ArrayList<Forme> objList;

Range sizeRange;
Range rotRange;
Range hueRange;
Range satRange;
Range briRange;
Range opaRange;

public void setup() {
  noLoop();

  /* Window configuration */
  size(600, 600);
  frame.setResizable(true);
  frame.addComponentListener(new ComponentAdapter() { 
   public void componentResized(ComponentEvent e) { 
     if(e.getSource()==frame) {
       Rectangle bounds = frame.getBounds();
       Insets insets = frame.getInsets();
       int width = bounds.width - insets.left - insets.right;
       int height = bounds.height - insets.top - insets.bottom;
       println("resize frame : " + width + " x " + height);
       randomize(width,height);
     } 
   } 
 });

  /* A list to retain the objects */
  objList = new ArrayList(nbObj);

  /* User Interface */
  controlP5 = new ControlP5(this);  
  controlP5.setAutoDraw(false);
  controlWindow = controlP5.addControlWindow("controlP5window",50,50,450,220);
  controlWindow.hideCoordinates();
  controlWindow.setBackground(color(40));
  controlWindow.setTitle("Param\u00e8tres DeadLeaves");

  /* Objects parameters */
  Slider s = controlP5.addSlider("nbObj",0,2048,20,20,350,10);
  s.setCaptionLabel("# Objects");
  s.moveTo(controlWindow);
  
  sizeRange = controlP5.addRange("size", 0, 512, 0, 256, 20, 40, 350, 10);
  sizeRange.moveTo(controlWindow);

  rotRange = controlP5.addRange("rotation", -360, 360, -180, 180, 20, 60, 350, 10);
  rotRange.setCaptionLabel("rotation (deg)");
  rotRange.moveTo(controlWindow);

  Radio r = controlP5.addRadio("type",20,90);
  r.moveTo(controlWindow);
  r.add("Ellipses",Forme.ELLIPSE);
  r.add("Rectangles",Forme.RECT);
  r.add("Circles",Forme.CIRCLE);
  r.add("Squares",Forme.SQUARE);
  r.add("Triangles",Forme.TRIANGLE);
  r.add("Stars",Forme.STAR);

  controlP5.addToggle("background",100,90,20,20).moveTo(controlWindow);
  controlP5.addToggle("seamless",160,90,20,20).moveTo(controlWindow);
  controlP5.addToggle("exponential",100,140,20,20).moveTo(controlWindow);
  controlP5.addToggle("sorted",160,140,20,20).moveTo(controlWindow);

  /* Color parameters */
  colorMode(HSB);  
  hueRange = controlP5.addRange("hue", 0, 255, 0, 255, 220, 90, 150, 10);
  hueRange.moveTo(controlWindow);
  satRange = controlP5.addRange("saturation", 0, 255, 0, 255, 220, 110, 150, 10);
  satRange.moveTo(controlWindow);
  briRange = controlP5.addRange("brightness", 0, 255, 0, 255, 220, 130, 150, 10);
  briRange.moveTo(controlWindow);
  opaRange = controlP5.addRange("opacity", 0, 255, 0, 255, 220, 150, 150, 10);
  opaRange.moveTo(controlWindow);

  /* Other controls */
  controlP5.addButton("randomize",0,20,190,80,19).moveTo(controlWindow);
  controlP5.addButton("export_PNG",0,120,190,80,19).moveTo(controlWindow);
  controlP5.addButton("export_PDF",0,220,190,80,19).moveTo(controlWindow);

  /* Create initial population */
  for(int i = 0; i < nbObj; i++) {
    addForme();
  }

}

public synchronized void draw() {
  beginPDF();
  if(background)
    background(0);
  ArrayList<Forme> tmp = objList;
  if(sorted) {
    tmp = new ArrayList(objList);
    Collections.sort(tmp);
  }
  for(Forme f : tmp) {
    if(seamless)
      f.drawSeamless();
    else
      f.draw();
  }
  println("draw " + objList.size() + " objects");
  endPDF();
}

public void addForme() {
  objList.add(
    new Forme(
      type,
      random(width),
      random(height), 
      randomFn(sizeRange.lowValue(), sizeRange.highValue()),
      randomFn(sizeRange.lowValue(), sizeRange.highValue()),
      color(
        random(hueRange.lowValue(), hueRange.highValue()),
        random(satRange.lowValue(), satRange.highValue()),
        random(briRange.lowValue(), briRange.highValue()),
        random(opaRange.lowValue(), opaRange.highValue())
      ),
      random(rotRange.lowValue(), rotRange.highValue())
    )
  );
}

public void randomize() {
  randomize(width, height);
}

public void randomize(int width, int height) {
  updateObjPosition(width, height);
  redraw();
}

public void updateObjNumber() {
    if(nbObj == objList.size())
      return;
    else if(nbObj < objList.size()) {
      objList.subList(nbObj,objList.size()).clear();
    } else {
      for(int i = objList.size(); i < nbObj; i++) {
        addForme();
      }
    }
}

public void updateObjSize() {
  println("begin change size");
  float min = sizeRange.lowValue();
  float max = sizeRange.highValue();
  for(Forme f: objList) {
    f.changeSize(
      randomFn(min,max),
      randomFn(min,max)
    );
  }
  println("end change size");
}

public void updateObjPosition(float width, float height) {
  for(Forme f: objList) {
    f.changePosition(random(width), random(height));
  }
}

public void updateObjType() {
  for(Forme f : objList) {
    f.changeType(type);
  }
}

public void updateObjRot() {
  for(Forme f : objList) {
    f.changeRot(random(rotRange.lowValue(), rotRange.highValue()));
  }
}

public void updateObjHue() {
  for(Forme f : objList) {
    f.changeHue(random(hueRange.lowValue(), hueRange.highValue()));
  }
}

public void updateObjSat() {
  for(Forme f : objList) {
    f.changeSat(random(satRange.lowValue(), satRange.highValue()));
  }
}

public void updateObjBri() {
  for(Forme f : objList) {
    f.changeBri(random(briRange.lowValue(), briRange.highValue()));
  }
}

public void updateObjOpa() {
  for(Forme f : objList) {
    f.changeOpa(random(opaRange.lowValue(), opaRange.highValue()));
  }
}

/* Keyboard */

public void keyPressed() {
  if(key=='r') {
    randomize();
  }
  if(key=='i') {
    if( controlWindow.isVisible() )
      controlWindow.hide();
    else
      controlWindow.show();
  }
  if(key == ' ') {
    export_PNG();
    export_PDF();
  }
}

/* Events */

public synchronized void controlEvent(ControlEvent theControlEvent) {
  String s = theControlEvent.controller().name();
  if(s.equals("size") || s.equals("exponential")) {
    println("change size");
    updateObjSize();
  } else if(s.equals("nbObj")) {
    println("change number of objects");
    updateObjNumber();
  } else if(s.equals("type")) {
    println("change type of objects");
    updateObjType();
  } else if(s.equals("hue")) {
    updateObjHue();
  } else if(s.equals("saturation")) {
    updateObjSat();
  } else if(s.equals("brightness")) {
    updateObjBri();
  } else if(s.equals("opacity")) {
    updateObjOpa();
  } else if(s.equals("rotation")) {
    updateObjRot();
  } else if(s.equals("export_PNG") || s.equals("export_PDF")) {
    return;
  }
  redraw();
}


/* Random distributions */
public float randomDisp(float r, float disp) {
  return random(min(0, r - disp), r + disp);
}

public float randomFn(float min, float max) {
  if(exponential)
    return randomExp(min, max);
  else
    return random(min, max);
}

public float randomExp(float min, float max) {
  float delta = max - min;
  return min + exp(random(-5,0))*delta;
}


/* Exportation */

public String date() {
    String s = Integer.toString(10000*year()+100*month()+day())
    + Integer.toString(10000*hour()+100*minute()+second())
      + "-" + Integer.toString(millis());
      return s;
}

public void export_PNG() {
  println("export PNG");
  saveFrame("deadLeaves-" + date() + ".png");
}
public void export_PDF() {
  println("export PDF");
  doSavePDF = true;
  redraw();
}

public void beginPDF() {
  if(doSavePDF)
    beginRecord(PDF, "deadLeaves-" + date() + ".pdf");
}

public void endPDF() {
  if(doSavePDF) {
    endRecord();
    doSavePDF=false;
  }
}

class Forme implements Comparable {

  final static int POINT = 0;
  final static int RECT = 1;
  final static int TRIANGLE = 2;
  final static int STAR = 3;
  final static int ELLIPSE = 4;
  final static int CIRCLE = 5;
  final static int SQUARE = 6;

  int t;      // type
  float x, y; // coordinates: X, Y
  float w, h; // size: width, height
  int c;    // color
  float r;    // rotation
  float ch, cs, cb, ca; // hue, saturation, brightness, alpha

  // Constructor
  Forme(int it, float ix, float iy, float iw, float ih, int ic, float ir) {
    t = it;
    x = ix;
    y = iy;
    w = iw;
    h = ih;
    c = ic;
    r = ir;
    
    //color components
    ch = hue(c);
    cs = saturation(c);
    cb = brightness(c);
    ca = alpha(c);
  }

  public void draw() {
    pushMatrix();
    translate(x,y);
    rotate(radians(r));
    c = color(ch, cs, cb, ca);
    fill(c); 
    noStroke();

    switch(t) {
    case ELLIPSE: 
      ellipse(0,0,w,h); 
      break;
    case CIRCLE:
      ellipse(0,0,w,w); 
      break;
    case SQUARE:
      rect(-w/2,-w/2,w,w);
      break;
    case RECT:
      rect(-w/2,-h/2,w,h);
      break;
    case TRIANGLE:
      triangle(0,-w,
      w*cos(PI*30/180),
      w*sin(PI*30/180),
      -w*cos(PI*30/180),
      w*sin(PI*30/180));
      break;
    case STAR:
      fill(c); 
      noStroke();
      star(0,0,w/2,w,5);
      break;
    default:
      stroke(c);
      point(x, y);
    }
    popMatrix();
  }

  private void star ( float x, float y, float radius_inner, float radius_outer, int spikes ) { 
    float r = 0.0f;
    float res = 360.0f/spikes;
    float res_half = res/2;

    beginShape();

    for ( float i = 0; i < 360; i+=res ) {
      r = -HALF_PI + radians( i );
      vertex( x + cos(r) * radius_outer, y + sin(r) * radius_outer );
      r = -HALF_PI + radians( i + res_half );
      vertex( x + cos(r) * radius_inner, y + sin(r) * radius_inner );
    } 

    endShape( CLOSE );
  } 


  public void drawSeamless() {
    translate(-width, -height);
    draw();
    translate(width, 0);
    draw();
    translate(width, 0);
    draw();
    translate(-2*width, height);
    draw();
    translate(width, 0);
    draw();
    translate(width, 0);
    draw();
    translate(-2*width, height);
    draw();
    translate(width, 0);
    draw();
    translate(width, 0);
    draw();
    translate(-width, -height);
  }

  public void changePosition(float ix, float iy) {
    x = ix;
    y = iy;
  }

  public void changeSize(float iw, float ih) {
    w = iw;
    h = ih;
  }

  public void changeType(int it) {
    t = it;
  }

  public void changeHue(float h) {
    ch = h;
  }

  public void changeSat(float s) {
    cs = s;
  }

  public void changeBri(float b) {
    cb = b;
  }

  public void changeOpa(float a) {
    ca = a;
  }

  public void changeRot(float ir) {
    r = ir;
  }

  public float size() {
    return min(w,h);
  }

  public int compareTo(Object f) {
    float size = size();
    float fsize = ((Forme) f).size();
    if(size < fsize) {
      return 1;
    } 
    else if (size == fsize) {
      return 0;
    } 
    else {
      return -1;
    }
  }
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#FFFFFF", "DeadLeaves_Generator" });
  }
}
